package ips.audio.coreaudio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.Control.Type;

public abstract class CoreAudioDataLine implements DataLine {

	public static final boolean DEBUG=false;
	
	public static final int DEFAULT_BUFFERSIZE_SECONDS=4;
	public static final int COREAUDIO_FLOAT_SAMPLESIZE=4;
	final static float KNOWN_SAMPLE_RATES[] = {
	    5512,
	    8000,
	    11025,
	    16000,
	    22050,
	    32000,
	    44100,
	    48000,
	    64000,
	    88200,
	    96000,
	    176400,
	    192000,
	};
	protected int writePos=0;
	protected int readPos=0;
	
	protected FloatBuffer ringBufferFloat;
	
	protected float maxValue;
	protected boolean signed=true;
	protected int sampleSize;
	
	protected boolean bigEndian=true;
	
	protected int physicalChannels;
	protected int channels;
	
	protected ByteBuffer dataline;
	protected ByteBuffer procId;
	private CoreAudioMixer mixer;
	private List<AudioFormat> supportedPhysicalFormats=null;
	private List<AudioFormat> supportedFormats=null;
	protected AudioFormat audioFormat;
	protected AudioFormat physicalAudioFormat;
	public List<AudioFormat> getSupportedFormats() {
		return supportedFormats;
	}
	protected DataLine.Info lineInfo = null;
	protected native void enumerateFormats(ByteBuffer streamID);
	protected native boolean start(ByteBuffer audioDeviceId);
	
	protected byte[] buffer;
	protected int bufferSize;
	protected int coreAudioBufferSize;
	protected Vector<LineListener> listeners;
	/**
	 * controls of this line (none implemented) 
	 */
	private Control[] controls = new Control[0];
	/**
	 * true if line is open
	 */
	protected boolean open = false;
	protected boolean active=false;
	
	public CoreAudioDataLine(CoreAudioMixer mixer,ByteBuffer dataline){
		super();
		this.mixer=mixer;
		this.dataline=dataline;
		listeners = new Vector<LineListener>();
		supportedPhysicalFormats=new ArrayList<AudioFormat>();
		supportedFormats=new ArrayList<AudioFormat>();
		enumerateFormats(dataline);
		
	}
	
	protected long coreAudioByteCountToJsByteCount(long coreAudioByteCount){
		return coreAudioByteCount*sampleSize*channels/COREAUDIO_FLOAT_SAMPLESIZE/physicalChannels;
	}
	protected long jsByteCountToCoreAudioByteCount(long jsBytes){
		return jsBytes*COREAUDIO_FLOAT_SAMPLESIZE*physicalChannels/sampleSize/channels;
	}
	protected native long getReadPos(ByteBuffer dataline);
	protected native long getWritePos(ByteBuffer dataline);
	
	protected AudioFormat getCorrespondingPhysicalAudioFormat(AudioFormat virtualFormat){
		float sampleRate=virtualFormat.getSampleRate();
		int sampleSizeBits=virtualFormat.getSampleSizeInBits();
		int channels=virtualFormat.getChannels();
		int maxSuppChannels=0;
		for(AudioFormat paf:supportedPhysicalFormats){
			int pafChs=paf.getChannels();
			if(pafChs>maxSuppChannels){
				maxSuppChannels=pafChs;
			}
		}
		for(AudioFormat paf:supportedPhysicalFormats){
			// look for "nearest" channel count
			for(int chs=channels;chs<=maxSuppChannels;chs++){
			if(paf.getSampleRate()==sampleRate &&
					paf.getSampleSizeInBits()==sampleSizeBits &&
					paf.getChannels()==chs){
				return paf;
			}
			}
		}
		
		return null;
	}
	
	protected void nextFormat(float sampleRate,int bits,int channels){
		AudioFormat paf=new AudioFormat(sampleRate,bits,channels,true,false);

		supportedPhysicalFormats.add(paf);
		if(DEBUG)System.out.println("fmt: "+paf);
		for(int chs=1;chs<=channels;chs++){
			AudioFormat afle=new AudioFormat(sampleRate,bits,chs,true,false);
			
			supportedFormats.add(afle);
			AudioFormat afbe=new AudioFormat(sampleRate,bits,chs,true,true);
			supportedFormats.add(afbe);
		}
	}
	protected void nextFormatRange(float sampleRateMin,float sampleRateMax,int bits,int channels){
		if(DEBUG)System.out.println("fmt (range): "+sampleRateMin+" - "+sampleRateMax);
		for(float knownSampleRate:KNOWN_SAMPLE_RATES){
			if(knownSampleRate>=sampleRateMin && knownSampleRate<=sampleRateMax){
				float sampleRate=knownSampleRate;
				AudioFormat paf=new AudioFormat(sampleRate,bits,channels,true,false);

				supportedPhysicalFormats.add(paf);
				if(DEBUG)System.out.println("fmt (range): "+paf);
				for(int chs=1;chs<=channels;chs++){
					AudioFormat afle=new AudioFormat(sampleRate,bits,chs,true,false);
					supportedFormats.add(afle);
					AudioFormat afbe=new AudioFormat(sampleRate,bits,chs,true,true);
					supportedFormats.add(afbe);
				}
			}
		}
	}
	
	
	
	@Override
	public void addLineListener(LineListener listener) {
		synchronized(listeners){
			if (listener != null && !listeners.contains(listener)) {
				listeners.addElement(listener);
			}
		}
	}
	
	
	protected void update(LineEvent le) {
		synchronized(listeners){
			for(LineListener ll:listeners){
				ll.update(le);
			}
		}
	}
	
	protected native ByteBuffer open(ByteBuffer streamId,float sampleRate,int sampleSizInBits,int channels, int arg1);
	
	public void open(AudioFormat arg0) throws LineUnavailableException {
		open(arg0,AudioSystem.NOT_SPECIFIED);
	}
	public void open(AudioFormat af, int arg1)
			throws LineUnavailableException {
		audioFormat=af;
		physicalAudioFormat=getCorrespondingPhysicalAudioFormat(audioFormat);
		if(physicalAudioFormat==null){
			throw new IllegalArgumentException("Audio format "+af+" not supported.");
		}
		channels=af.getChannels();
		physicalChannels=physicalAudioFormat.getChannels();
		int frameSize=af.getFrameSize();
		sampleSize = frameSize / channels;
//		int physicalFrameSize=physicalChannels*sampleSize;
		if(arg1==AudioSystem.NOT_SPECIFIED){
			// four seconds
			bufferSize=(int)audioFormat.getSampleRate()*DEFAULT_BUFFERSIZE_SECONDS*frameSize;
		}else{
			bufferSize=arg1;
		}
		
		coreAudioBufferSize=(int)jsByteCountToCoreAudioByteCount((long)bufferSize);
		ByteBuffer ringBuffer=open(dataline,af.getSampleRate(),af.getSampleSizeInBits(),physicalChannels,coreAudioBufferSize);
		if(ringBuffer==null){
			throw new LineUnavailableException("Could not open core audio line."); 
		}
		ringBuffer.order(ByteOrder.LITTLE_ENDIAN);
		ringBufferFloat=ringBuffer.asFloatBuffer();
		

		//			buffer=new byte[bufferSize];
		int resolutionBits=af.getSampleSizeInBits();
		if (resolutionBits > 32)
			throw new IllegalArgumentException("sample sizes above 32 bit not supported");

		resolutionBits = af.getSampleSizeInBits();
		float pp = (float) Math.pow(2.0, (double) resolutionBits);
		maxValue = pp / 2;
		bigEndian = af.isBigEndian();
		open();
	}

	private native void close(ByteBuffer dl);
	@Override
	public void close() {
		if(open){
			close(dataline);
			open = false;
			update(new LineEvent(this, LineEvent.Type.CLOSE, getLongFramePosition()));
		}
	}

	@Override
	public Control getControl(Type control) {
		return null;
	}

	@Override
	public Control[] getControls() {
		return controls;
	}

	@Override
	public javax.sound.sampled.Line.Info getLineInfo() {
		return lineInfo;
	}

	@Override
	public boolean isControlSupported(Type control) {
		return false;
	}

	@Override
	public boolean isOpen() {
		return open;
	}

	@Override
	public void open() throws LineUnavailableException {
		if (open)
			return;
		open = true;
		update(new LineEvent(this, LineEvent.Type.OPEN, 0));
	}

	@Override
	public void removeLineListener(LineListener listener) {
		synchronized(listeners){
			if (listener != null) {
				listeners.removeElement(listener);
			}
		}
	}

	

	@Override
	public abstract void drain();

	@Override
	public abstract void flush();

	@Override
	public int getBufferSize() {
		
		return bufferSize;
	}

	@Override
	public AudioFormat getFormat() {
		
		return audioFormat;
	}
	
	

	@Override
	public int getFramePosition() {
		return (int)getLongFramePosition();
	}

	@Override
	public float getLevel() {
		return -1;
	}

	@Override
	public abstract long getLongFramePosition();
		

	@Override
	public long getMicrosecondPosition() {
		long framePos=getLongFramePosition();
		double sPos = ((double) framePos * 1000000);
		double msPos = sPos / audioFormat.getFrameRate();
		if (DEBUG)
			System.out.println(
				framePos + " " + audioFormat.getFrameRate() + " " + sPos + " " + msPos);

		return (long) msPos;
	}

	@Override
	public boolean isActive() {
		return active;
	}

	protected native boolean isRunning(ByteBuffer dataline);
	@Override
	public boolean isRunning() {
		return isRunning(dataline);
	}

	@Override
	public void start() {
		if(!active){
			if(start(dataline)){
				active=true;
				update(new LineEvent(this, LineEvent.Type.START, getLongFramePosition()));
			}
		}
	}

	@Override
	public void stop() {
		if(active){
			if(stop(dataline)){
				active=false;
				update(new LineEvent(this, LineEvent.Type.STOP, getLongFramePosition()));
			}
		}
	}
	private native boolean stop(ByteBuffer dataline);

}
