
package ips.audio.coreaudio;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;

/**
 * Implements the mixer interface for CoreAudio alternative JavaSound (AJS) implementation
 * 
 * @author Klaus Jaensch, Institute of Phonetics, klausj@phonetik.uni-muenchen.de
 * 
 */

public class CoreAudioMixer implements Mixer {

	final static boolean DEBUG = false;
	
	private CoreAudioMixerInfo minfo;
	private ByteBuffer audioDeviceID;
	public ByteBuffer getAudioDeviceID() {
		return audioDeviceID;
	}
	boolean open = false;

	private Line.Info linfo;
	
	private DataLine.Info slDataLineInfo = null;

	private Control[] controls = new Control[0];
	private List<CoreAudioTargetDataLine> tdls = null;
	private List<CoreAudioSourceDataLine> sdls = null;
	private Vector<AudioFormat> tmpSupportedTargetFormats;
	private AudioFormat[] suppTargetFormats;

	private static Mixer.Info openedMixerInfo = null;
	 private native String getName(ByteBuffer audioDeviceId);
	private native void enumerateCaptureStreams(ByteBuffer audioDeviceId);
	private native void enumeratePlaybackStreams(ByteBuffer audioDeviceId);
	
	public CoreAudioMixer(ByteBuffer audioDeviceID,boolean useAJSSuffix){
		super();
		this.audioDeviceID=audioDeviceID;
		String deviceName=getName(audioDeviceID);
		this.minfo=new CoreAudioMixerInfo(deviceName,useAJSSuffix);
		tdls=new ArrayList<CoreAudioTargetDataLine>();
		sdls=new ArrayList<CoreAudioSourceDataLine>();
		enumerateCaptureStreams(audioDeviceID);
		enumeratePlaybackStreams(audioDeviceID);
		linfo=new Line.Info(getClass());
		
		if(DEBUG)System.out.println("Constructed: "+this);
		
	}
	
	protected void nextCaptureStream(ByteBuffer dataline){
		tdls.add(new CoreAudioTargetDataLine(this,dataline));	
	}
	
	protected void nextPlaybackStream(ByteBuffer dataline){
		sdls.add(new CoreAudioSourceDataLine(this,dataline));	
	}
	
	public String toString(){
		return "CoreAudio mixer : "+minfo;
	}
	public void open() throws LineUnavailableException {
		if (! open)
			return;
		
	
		openedMixerInfo = minfo;
		open = true;
	}

	/* Line interface */

	public void addLineListener(LineListener listener) {
	}
	public synchronized void close() {
		open = false;
		openedMixerInfo = null;
		if (DEBUG)
			System.out.println("DSMixer closed (released).");
	}
	public Control getControl(Control.Type control) {
		return null;
	}
	public Control[] getControls() {
		return controls;
	}
	public Line.Info getLineInfo() {
		return linfo;
	}
	public boolean isControlSupported(Control.Type control) {
		return false;
	}
	public boolean isOpen() {
		return open;
	}

	public void removeLineListener(LineListener listener) {
	}

	/* Mixer interface */
	public Line getLine(Line.Info info) throws LineUnavailableException {
		for(CoreAudioTargetDataLine tdl:tdls){
			Line.Info tdlInfo=tdl.getLineInfo();
			if (tdlInfo != null && info.matches(tdlInfo)){
				return tdl;
			}
		}
		for(CoreAudioSourceDataLine sdl:sdls){
			Line.Info dlInfo=sdl.getLineInfo();
			if (dlInfo != null && info.matches(dlInfo)){
				return sdl;
			}
		}
		throw new IllegalArgumentException();
	}

	public int getMaxLines(Line.Info info) {
		// TODO for now we only allow one line
		return 1;
	}

	public Mixer.Info getMixerInfo() {
		return minfo;
	}

	public Line.Info[] getSourceLineInfo() {
		List<Line.Info> matchingInfos=new ArrayList<Line.Info>();

		for(CoreAudioSourceDataLine sdl:sdls){
			Line.Info dlInfo=sdl.getLineInfo();
			matchingInfos.add(dlInfo);
		}
		return matchingInfos.toArray(new Line.Info[0]);
	}

	public Line.Info[] getSourceLineInfo(Line.Info info) {
		if(info!=null){
			List<Line.Info> matchingInfos=new ArrayList<Line.Info>();

			for(CoreAudioSourceDataLine sdl:sdls){
				Line.Info dlInfo=sdl.getLineInfo();
				if (dlInfo != null && info.matches(dlInfo)){
					matchingInfos.add(dlInfo);
				}
			}
			return matchingInfos.toArray(new Line.Info[0]);
		}else{
			return new Line.Info[0];
		}
	}

	public Line[] getSourceLines() {
		return sdls.toArray(new Line[0]);
	}
	
	private native void getCaps(byte[] np,boolean source);
	
	
	public Line.Info[] getTargetLineInfo() {
		List<Line.Info> matchingInfos=new ArrayList<Line.Info>();

		for(CoreAudioTargetDataLine tdl:tdls){
			Line.Info tdlInfo=tdl.getLineInfo();
			matchingInfos.add(tdlInfo);
		}
		return matchingInfos.toArray(new Line.Info[0]);
	}

	
	public Line.Info[] getTargetLineInfo(Line.Info info) {

		if(info!=null){
			List<Line.Info> matchingInfos=new ArrayList<Line.Info>();

			for(CoreAudioTargetDataLine tdl:tdls){
				Line.Info tdlInfo=tdl.getLineInfo();
				if (tdlInfo != null && info.matches(tdlInfo)){
					matchingInfos.add(tdlInfo);
				}
			}
			return matchingInfos.toArray(new Line.Info[0]);
		}else{
			return new Line.Info[0];
		}
	}

	public Line[] getTargetLines() {
		return tdls.toArray(new Line[0]);
	}

	public boolean isLineSupported(Line.Info info) {
		
		if (info.matches(linfo))
			return true;
		
		if(TargetDataLine.class.isAssignableFrom(info.getLineClass())){
			Line.Info[] lInfos=getTargetLineInfo(info);
			return(lInfos!=null && lInfos.length>0);
		}else if(SourceDataLine.class.isAssignableFrom(info.getLineClass())){
			Line.Info[] lInfos=getSourceLineInfo(info);
			return(lInfos!=null && lInfos.length>0);
		}
		return false;
	}

	public boolean isSynchronizationSupported(
		Line[] lines,
		boolean maintainSync) {
		return false;
	}

	public void synchronize(Line[] lines, boolean maintainSync) {
	}

	public void unsynchronize(Line[] lines) {
	}
	
//	protected void finalize() throws Throwable{
//		System.out.println("Releasing "+this);
//		
//		super.finalize();
//		if(DEBUG)System.out.println("Mixer finalized");
//	}

}
